/*! \file XMLParsingMQL4.mq4
    \brief   .
    \author <a href="http://www.mql4.com/ru/users/BenMond">BenMond</a>
    
     //,      .
\code
#include <XmlParsingMQL4.mqh> //  -     
\endcode

*/

#property copyright "Copyright  2011, Ben"
#property link      "http://www.MyFXResearch.com"

#property library

#include <XmlParsingMQL4.mqh> //  -     

#define XTW_Handle    0       
#define XTW_TagOpened 1
#define XTW_Size      2

static int    XTR_BufferSize            = 1000;
static string StrEnd                    = "\r\n";
static string EndOfXmlDeclarationStr    = "?>";
static string StartOfXmlCommentStr      = "<!--";
static string EndOfXmlCommentStr        = "-->";

/** \addtogroup SAX_Write
 *  @{
 */

///   FileName  .     - ,   -  .       -1   .
int XmlTextWriter(int& XmlTextWriter[],   ///< [in,out] XmlTextWriter Object
                  string FileName         ///< [in]    
                  ) {

   ArrayResize(XmlTextWriter, XTW_Size);

   XmlTextWriter[XTW_Handle] = FileOpen(FileName, FILE_BIN|FILE_WRITE);
   XmlTextWriter[XTW_TagOpened] = false;
   
   return (XmlTextWriter[XTW_Handle]);
}

///    Xml  <?Xml version="1.0" encoding="windows-1251" ?> . 
void WriteStartDocument(int& XmlTextWriter[], ///< [in,out] XmlTextWriter Object
                        string encoding       ///< [in] .  "windows-1251"
                        ) {
   string StartStr = "<?xml version=\"1.0\" encoding=\""+encoding+"\" ?>"+StrEnd;
   FileWriteString(XmlTextWriter[XTW_Handle], StartStr, StringLen(StartStr));
}

///      Name. <Name>
void WriteStartElement(int& XmlTextWriter[], ///< [in,out] XmlTextWriter Object
                        string Name          ///< [in]  . 
                        ) {
   string StartElement = "";
   if(XmlTextWriter[XTW_TagOpened] == true) StartElement = ">"+StrEnd; 
   StartElement = StartElement + "<"+Name;
   XmlTextWriter[XTW_TagOpened] = true;
   FileWriteString(XmlTextWriter[XTW_Handle], StartElement, StringLen(StartElement));
}

///   attrName   value. <Name attrName="value">
void WriteAttributeString(int& XmlTextWriter[], ///< [in,out] XmlTextWriter Object
                           string attrName,     ///< [in]  . 
                           string value         ///< [in]  .
                           ) {
   string AttributeString = " "+attrName+"=\""+TranslateToEntity(value)+"\"";
   XmlTextWriter[XTW_TagOpened] = true;
   FileWriteString(XmlTextWriter[XTW_Handle], AttributeString, StringLen(AttributeString));

}

///     text. <Name>text</Name>
void WriteString(int& XmlTextWriter[],    ///< [in,out] XmlTextWriter Object
                  string text             ///< [in]  . 
                  ) {
   string String = ">"+TranslateToEntity(text);
   
   XmlTextWriter[XTW_TagOpened] = false; 
   
   FileWriteString(XmlTextWriter[XTW_Handle], String, StringLen(String));
}

///     Name   value. <Name>value</Name>
void WriteElementString(int& XmlTextWriter[], ///< [in,out] XmlTextWriter Object
                        string Name,          ///< [in]  . 
                        string value          ///< [in]  .
                        ) {                    
   WriteStartElement(XmlTextWriter, Name);
   WriteString(XmlTextWriter, value);
   WriteEndElement(XmlTextWriter, Name);
}

///   <!-- text --> 
void WriteComment(int& XmlTextWriter[],   ///< [in,out] XmlTextWriter Object
                  string text             ///< [in]  .
                  ) {
   string String = "";
   if(XmlTextWriter[XTW_TagOpened] == true) {
      String = ">"+StrEnd; 
      XmlTextWriter[XTW_TagOpened] = false;
   }
   String = String + StartOfXmlCommentStr + " "+text+" " + EndOfXmlCommentStr + StrEnd;
   
   FileWriteString(XmlTextWriter[XTW_Handle], String, StringLen(String));
}

///     Name. </Name>
void WriteEndElement(int& XmlTextWriter[],   ///< [in,out] XmlTextWriter Object
                     string Name             ///< [in]  . 
                     ) {
   string String = "";
   
   if(XmlTextWriter[XTW_TagOpened] == true) {
      String = " />"; 
      XmlTextWriter[XTW_TagOpened] = false;
   }
   else String = XmlDocEndElement(Name);
   
   String = String+StrEnd;
   
   FileWriteString(XmlTextWriter[XTW_Handle], String, StringLen(String));
}

///  . 
void WriteEndDocument(int& XmlTextWriter[]   ///< [in,out] XmlTextWriter Object
                     ) {
   FileClose(XmlTextWriter[XTW_Handle]);
}

/** @}*/

#define XTR_Handle    0
#define XTR_LastPos   1
#define XTR_Size      2

/** @addtogroup SAX_Read
 *  @{
 */

///   FileName  .   XmlTextReader.  true   ,  false.
bool XmlTextReader(int& XmlTextReader[], ///< [in,out] XmlTextReader Object
                     string FileName     ///< [in]  
                     ) {
   ArrayResize(XmlTextReader, XTR_Size);

   XmlTextReader[XTR_Handle] = FileOpen(FileName, FILE_BIN|FILE_READ);
   bool Opened = (XmlTextReader[XTR_Handle]!=-1);
   if(Opened) {
      string buffer = ReadBuffer(XmlTextReader);
      ReadStrUpTo(XmlTextReader, buffer, EndOfXmlDeclarationStr);
   }
   return (Opened);
}

///         .       .
string XmlRead(int& XmlTextReader[]       ///< [in,out] XmlTextReader Object
               ) {
   string buffer = ReadBuffer(XmlTextReader);
   string result;
   if(buffer=="") {
      result = buffer;
   } else {
      if(XmlDocIsComment(buffer)) {
         result = StringSubstr(ReadStrUpTo(XmlTextReader, buffer, EndOfXmlCommentStr),4);
      } else if(XmlDocIsEndElement(buffer)) {
         result = ReadStrUpTo(XmlTextReader, buffer, ">");
      } else if(XmlDocIsElement(buffer)) {
         result = ReadStrUpTo(XmlTextReader, buffer, ">");
      } else if(XmlDocIsText(buffer)) {
         result = TranslateFromEntity(ReadStrBefore(XmlTextReader, buffer, "<"));
      }
   }
   return (result);
}

///  XmlTextReader.
void XmlClose(int& XmlTextReader[]     ///< [in,out] XmlTextReader Object
               ) {
   FileClose(XmlTextReader[XTR_Handle]);
}

/** @}*/

/** @addtogroup GroupDOM
 *  @{
 */

///  Xml  FileName.  .
string XmlDocument(string FileName  ///< [in]  
                   ) {
   string doc="";
   int Handle = FileOpen(FileName, FILE_BIN|FILE_READ);
   if(Handle!=-1) {
      doc = FileReadString(Handle,  FileSize(Handle) );
      GetStrUpTo(doc, EndOfXmlDeclarationStr);
   }
   return (doc);
}

///    Node   aChildNodes.
void XmlDocChildNodes(string Node,              ///< [in] Xml 
                      string &aChildNodes[]     ///< [out]    
                      ) {
   string Tag = XmlDocGetFirstTag(Node);
   
   ArrayResize(aChildNodes,0);
   if(XmlDocIsElement(Tag)){
      if(!XmlDocIsClosedTag(Tag)) {
         string name = XmlDocName(Tag);
         while((XmlDocName(Node)!=name)&&(Node!="")) {
            string ChildNode = XmlDocGetFirstNode(Node);
            AddNewStr(aChildNodes, ChildNode);
         }
      }
   }
}

///    Child
string XmlDocAppendChild(string Node,  /// [in] Xml 
                         string Child  /// [in]  
                         ) {
   string aChilds[];
   XmlDocChildNodes(Node, aChilds);
   AddNewStr(aChilds, Child);
   Node = XmlDocGetFirstTag(Node);
   for(int i=0;i<ArraySize(aChilds);i++) {
      Node = Node+aChilds[i];
   }
   Node = Node+XmlDocEndElement(XmlDocName(Node));
   return (Node);
}

///      Name   aNodes.
void XmlDocGetElementsByTagName(string doc,           /// [in] Xml 
                                string &aNodes[],     /// [out]   
                                string Name           /// [in]  
                                ) {
   string Tag;
   
   ArrayResize(aNodes,0);
   while(doc!="") {
      Tag = XmlDocGetFirstTag(doc);
      if(XmlDocName(Tag)==Name) {
         string Node = Tag;
         if(!XmlDocIsClosedTag(Node)) {
            Tag = "";
            while((doc!="")&&(XmlDocName(Tag)!=Name)) {
               Tag = XmlDocGetFirstTag(doc);
               Node = Node+Tag;
            }
         }
         AddNewStr(aNodes, Node);
      }
   }
}

///  doc 
string XmlDocFormat(string doc   /// [in] Xml 
                     ) {
   string rest = doc;
   string Formated="";
   int    nSpaces = 0;
   
   string Tag = XmlDocGetFirstTag(rest);
   if(XmlDocIsElement(Tag)) {
      Formated = Formated + Tag;
      string name = XmlDocName(Tag);
      while (rest!="") {
         Tag = XmlDocGetFirstTag(rest);
         if(XmlDocIsText(Tag)) {
            Formated = Formated + Tag;
         } else if(XmlDocIsComment(Tag)) {
            Formated = Formated + StrEnd + Spaces(nSpaces) + Tag;
         } else if(XmlDocIsClosedTag(Tag)) {
            nSpaces++;
            Formated = Formated + StrEnd + Spaces(nSpaces) + Tag;
            nSpaces--;
            name = "";
         } else if(XmlDocIsElement(Tag)) {
            nSpaces++;
            Formated = Formated + StrEnd + Spaces(nSpaces) + Tag;
            name = XmlDocName(Tag);
         } else if(XmlDocIsEndElement(Tag)&&(XmlDocName(Tag)==name)) {
            nSpaces--;
            Formated = Formated + Tag;
            name = "";
         } else if(XmlDocIsEndElement(Tag)&&(XmlDocName(Tag)!=name)) {
            Formated = Formated + StrEnd + Spaces(nSpaces) + Tag;
            name = "";
            nSpaces--;
         } else {
            Formated = Formated + StrEnd + Spaces(nSpaces) + Tag;
         }
      }
   } else {
      Formated = Formated + Tag + StrEnd;
   }
   
   return (Formated);
}

///  doc  OutPutFileName
void XmlDocSave(string doc,            /// [in] Xml 
                string OutPutFileName, ///  
                string encoding        ///  "windows-1251"
                ) {
   int Handle = FileOpen(OutPutFileName, FILE_BIN|FILE_WRITE);
   if(Handle!=-1) {
      if(!EqualBegin(doc, "<?")) {
         string StartStr = "<?xml version=\"1.0\" encoding=\""+encoding+"\" ?>"+StrEnd;
         FileWriteString(Handle, StartStr, StringLen(StartStr));
      }
      doc = XmlDocFormat(doc);
      FileWriteString(Handle, doc, StringLen(doc));
   }
}  

///     Name   <Name></Name>
string XmlDocCreateElement(string Name /// [in]  
                           ) {
   return("<"+Name+"></"+Name+">");
}

///     <Name>Text</Name>
string XmlDocText(string Node                   ///< [in] 
                  ) {
   string Tag = XmlDocGetFirstTag(Node);
   string text = XmlDocGetFirstTag(Node);
   return (TranslateFromEntity(text));
}

///    text  <Name>text</Name>
string XmlDocSetElementText(string Node, /// [in] 
                            string text  /// [in] 
                            ) {
   string tag = XmlDocGetFirstTag(Node);      
   string name = XmlDocName(tag);
   Node = tag+TranslateToEntity(text)+XmlDocEndElement(name);
   return (Node);
}

///   AttrName   value
string XmlDocAppendAttribute(string Node,       /// [in] 
                             string AttrName,   /// [in]  
                             string value       /// [in]  
                             ) {
   string Tag = XmlDocGetFirstTag(Node);
   if(XmlDocIsElement(Tag)) {   
      string Attribute = AttrName+"=\""+TranslateToEntity(value)+"\" ";
      bool closed = XmlDocIsClosedTag(Tag);
   
      if(closed) {
         Tag = StringSubstr(Tag, 0, StringLen(Tag)-2);
      } else {
         Tag = StringSubstr(Tag, 0, StringLen(Tag)-1);
      }

      if(!EqualEnd(Tag," ")) Tag = Tag+" ";

      if(closed) {
         Tag = Tag+Attribute+"/>";
      } else {
         Tag = Tag+Attribute+">";
      }
   }
      
   Node = Tag + Node;
   
   return (Node);
}

/** @}*/

/** @addtogroup TYPE
 *  @{
 */

///      <item>
bool XmlDocIsElement(string Tag  /// [in] 
                     ) {
   return (EqualBegin(Tag,"<")&&(!XmlDocIsEndElement(Tag))&&(!XmlDocIsComment(Tag)));
}

///      </item>
bool XmlDocIsEndElement(string Tag  /// [in] 
                        ) {
   return (EqualBegin(Tag,"</"));
}

///      
bool XmlDocIsText(string Tag  /// [in] 
                  ) {
   return (!EqualBegin(Tag,"<"));
}

///     <!-- My Comment -->
bool XmlDocIsComment(string Tag  /// [in] 
                     ) {
   return (EqualBegin(Tag, StartOfXmlCommentStr));
}

///   Tag . <item />
bool XmlDocIsClosedTag(string Tag   /// [in] 
                       ) {
   return (EqualBegin(Tag,"<")&&EqualEnd(Tag,"/>"));
}

/** @}*/

/** @addtogroup COMMON
 *  @{
 */

///    . <Name>
string XmlDocName(string Node       ///< [in] 
                  ) {
   string Name = "";   

   if     (EqualBegin(Node,"</")) Name = StringSubstr(Node,2);
   else if(EqualBegin(Node,"<" )) Name = StringSubstr(Node,1);

   if(Name!="") {
      int bufferlen = StringLen (Name);
      int SpacePos  = StringFind(Name, " ");
      int SlashPos  = StringFind(Name, "/");
      int EndPos    = StringFind(Name, ">");
         
      if(SpacePos==-1) SpacePos = bufferlen;
      if(SlashPos==-1) SlashPos = bufferlen;
      if(EndPos  ==-1) EndPos   = bufferlen;
      
      int LastPos = MathMin(MathMin(EndPos,SlashPos),SpacePos);
   
      Name = StringSubstr(Name,0,LastPos);
   }
   
   return(Name);
}

///   AttrName.
string XmlDocAttribute(string Node,       ///< [in] 
                       string AttrName    ///< [in]  
                       ) {
   string value = "";
   int NamePos = StringFind(Node, " "+AttrName+"=");
   if(NamePos!=-1) {
      value = StringSubstr(Node, NamePos+StringLen(AttrName)+2);
      int EndPos = StringLen(value);
      string firstChar = StringSubstr(value,0,1);
      int firstPos  = 0;
      if(   false
         ||(firstChar=="\"") 
         ||(firstChar=="\'") 
         ){
         EndPos = StringFind(value, firstChar,1);
         firstPos = 1;
      } else {
         EndPos = StringFind(value, " ",1);
      }
      value = StringSubstr(value, firstPos, EndPos-firstPos);
   }
   return (TranslateFromEntity(value));
}

/** @}*/

///  nSpaces 
string Spaces(int nSpaces  /// [in]  
               ) {
   string Spaces = "";
   for(int i=0; i<nSpaces; i++) {
      Spaces = Spaces + "  ";
   }
   return (Spaces);
}

///   .
bool EqualBegin(string str1, /// [in]  
                string str2  /// [in]  
                ) {
   int minLen = MathMin(StringLen(str1),StringLen(str2));
   return (StringSubstr(str1,0,minLen)==StringSubstr(str2,0,minLen));
}

///   
bool EqualEnd(string str1, /// [in]  
              string str2  /// [in]  
              ) {
   int Len1   = StringLen(str1);
   int Len2   = StringLen(str2);
   int minLen = MathMin(Len1,Len2);
   return (StringSubstr(str1,Len1-minLen,minLen)==StringSubstr(str2,Len2-minLen,minLen));
}

///       value
void AddNewStr(string& aStr[], /// [in, out]  
               string value    /// [in]  
               ) {
   ArrayResize(aStr, ArraySize(aStr)+1);
   aStr[ArraySize(aStr)-1] = value;
}

///  & amp;  &, & lt;  <, & gt;  >, & apos;  ', & quot;  "
string TranslateFromEntity(string Str  /// [in]  
                           ) {
   Str = StringReplace(Str, "&amp;" , "&" );
   Str = StringReplace(Str, "&lt;"  , "<" );
   Str = StringReplace(Str, "&gt;"  , ">" );
   Str = StringReplace(Str, "&apos;", "\'");
   Str = StringReplace(Str, "&quot;", "\"");
   return (Str);
}

///  &  & amp; , <  & lt; , >  & gt; , '  & apos; , "  & quot;
string TranslateToEntity(string Str    /// [in]  
                         ) {

   Str = StringReplace(Str, "&" , "&amp;" );
   Str = StringReplace(Str, "<" , "&lt;"  );
   Str = StringReplace(Str, ">" , "&gt;"  );
   Str = StringReplace(Str, "\'", "&apos;");
   Str = StringReplace(Str, "\"", "&quot;");

   return (Str);
}

///   Str   Search  Replace
string StringReplace(string Str,       /// [in]  
                     string Search,    /// [in]  
                     string Replace    /// [in]   
                     ) {
   int i = 0, next;
   bool hasSearchString = true;
   string NewStr = "";
   
   while (hasSearchString) {
      next = StringFind(Str, Search, i);
      if(next==-1) {
         NewStr = NewStr + StringSubstr(Str, i);
         hasSearchString = false;
      } else {
         if((next-i)>0) NewStr = NewStr + StringSubstr(Str, i, next-i);
         NewStr = NewStr + Replace;
         i = next+StringLen(Search);
      }
   }
   
   return (NewStr);
}

///     StopStr
string ReadStrBefore(int& XmlTextReader[], ///< [in,out] XmlTextReader Object
                     string buffer,        ///< [in]  
                     string StopStr        ///< [in]  
                     ) {
   string result;
   int EndPos = StringFind(buffer, StopStr);
   if(EndPos==-1) {
      EndPos = StringLen(buffer);
   }
   result = StringSubstr(buffer, 0, EndPos);
   int NextNodePos = EndPos+StringLen(StopStr)-1;
   SetPos(XmlTextReader, NextNodePos);
   return (result);
}

///     StopStr    .
string ReadStrUpTo(int& XmlTextReader[],  ///< [in,out] XmlTextReader Object
                   string buffer,         ///< [in]  
                   string StopStr         ///< [in]  
                   ) {
   string result;
   int EndPos = StringFind(buffer, StopStr);
   if(EndPos==-1) {
      EndPos = StringLen(buffer);
   }
   int NextNodePos = EndPos+StringLen(StopStr);
   result = StringSubstr(buffer, 0, EndPos);
   int nSpaces = CalcSpaces(buffer, NextNodePos);
   NextNodePos += nSpaces;
   SetPos(XmlTextReader, NextNodePos);
   return (result);
}

///   
string ReadBuffer(int& XmlTextReader[]       ///< [in,out] XmlTextReader Object
                  ) {
   XmlTextReader[XTR_LastPos] = FileTell(XmlTextReader[XTR_Handle]);
   return (FileReadString(XmlTextReader[XTR_Handle], XTR_BufferSize));
}

//      
string SetPos(int& XmlTextReader[],    ///< [in,out] XmlTextReader Object
              int Pos                  ///< [in]    
              ) {
   return (FileSeek(XmlTextReader[XTR_Handle], XmlTextReader[XTR_LastPos]+Pos, SEEK_SET));
}

///   
int CalcSpaces(string buffer, /// [in]  
               int nStart     /// [in]  
               ){
   int nSpaces = 0;
   for(int iPos = nStart; iPos<StringLen(buffer); iPos++) {
      if(isSpace(StringSubstr(buffer, iPos,1))) {
         nSpaces++;
      } else {
         break;
      }
   }
   return(nSpaces);
}

///   sChar 
bool isSpace(string sChar  /// [in]  
             ) {
   bool isChar = false;
   static string aSpaces[] = {" ","\r","\n","\t"};
   for(int i=0;i<ArraySize(aSpaces);i++) {
      if(sChar==aSpaces[i]) {
         isChar = true;
         break;
      }
   }
   
   return(isChar);
}

///     StopStr 
string GetStrUpTo(string &doc,      /// [in,out]  
                  string StopStr    /// [in]  
                  ) {
   string result;
   int EndPos = StringFind(doc, StopStr);
   if(EndPos==-1) {
      EndPos = StringLen(doc);
   }
   int NextNodePos = EndPos+StringLen(StopStr);
   result = StringSubstr(doc, 0, NextNodePos);
   int nSpaces = CalcSpaces(doc, NextNodePos);
   NextNodePos += nSpaces;
   doc    = StringSubstr(doc, NextNodePos);
   return (result);
}

///     StopStr
string GetStrBefore(string &doc,    /// [in,out]  
                    string StopStr  /// [in]  
                    ) {
   string result;
   int EndPos = StringFind(doc, StopStr);
   if(EndPos==-1) {
      EndPos = StringLen(doc);
   }
   result = StringSubstr(doc, 0, EndPos);
   int NextNodePos = EndPos+StringLen(StopStr)-1;
   doc = StringSubstr(doc, NextNodePos);
   return (result);
}

///    .    doc.
string XmlDocGetFirstNode(string& doc  /// [in,out]  
                          ) {
   string Node = "";
   string Tag = XmlDocGetFirstTag(doc);
   
   if(XmlDocIsComment(Tag)) {
      Node = Tag;
   } else if(XmlDocIsEndElement(Tag)) {
      Node = Tag;
   } else if(XmlDocIsElement(Tag)){
      Node = Tag;
      if(XmlDocIsClosedTag(Tag)) {
         Node = Node;
      } else {
         string name = XmlDocName(Tag);
         bool IsFirstTag = true;
         while((XmlDocName(doc)!=name)&&(doc!="")) {
            Tag = XmlDocGetFirstTag(doc);
            Node = Node + Tag;
         }
      }
   } else {
      Node = Tag;
   }
   return (Node);
}

///        doc
string XmlDocGetFirstTag(string &doc   /// [in,out]  
                        ) {
   string Tag = "";
   if(EqualBegin(doc, StartOfXmlCommentStr)) {
      Tag = GetStrUpTo(doc, EndOfXmlCommentStr);
   } else if(EqualBegin(doc, "<" )){
      Tag = GetStrUpTo(doc, ">");
   } else {
      Tag = GetStrBefore(doc, "<");
   }
   return (Tag);
}

///    </Name>
string XmlDocEndElement(string Name /// [in]  
                        ) {
   return ("</"+Name+">");
}


